import pandas as pd
meta_data_path = "/ocean/projects/asc170022p/shared/Data/ukBiobank/meta_data_files/"
fs_combat_cols = ['eid', '53-2.0', '34-0.0', '31-0.0', '54-2.0']


class harmonize_fs_data:
    def __init__(self):
        return

    @staticmethod
    def get_subject_ids():
        """
        A utility function which lets user fetch a list of subject ids whose free-surfer data is available on PSC
        Parameters:
        No parameter required
        Returns:
        subject list: A python list object

        """
        with open(meta_data_path+'UKB_fs_folder_paths_march_2022.csv') as f:
            folder_paths = f.read().splitlines()

        fs_subject_list = []
        for path in folder_paths:
            fs_subject_list.append(path.split("/")[-3])

        return list(set(fs_subject_list))

    @staticmethod
    def ukb_subject_sex(sex):
        if sex == 0.0:
            return "F"
        else:
            return "M"

    @staticmethod
    def ukb_subject_age(year):
        return int(2022 - year)

    @staticmethod
    def ukb_subject_age_at_scan(row):
        return int(row['Date_of_Brain_Scan'].year - row['Year_of_Birth'])

    @staticmethod
    def parse_subject_id_from_path(path):
        return path.split("/")[-3]

    @staticmethod
    def get_center(center):
        if center == 11025.0:
            return "Cheadle"

        elif center == 11026.0:
            return "Reading"

        elif center == 11027.0:
            return "Newcastle"

    def get_demographics_data(self, subject_list=None):
        """
        The primary function of this module which lets user provide a subject list. Returned is a pandas dataframe which
        has metadata and freesurfer features. If no subject list is provided all subjects are considered
        Parameters:
        A python list of subjects whose freesurfer and metadata features is required
        Returns:
        subject list: A pandas dataframe which has all the features from freesurfer like cortical measurements and
        subcortical volume numbers
        """
        if subject_list is None:
            fs_subject_list = harmonize_fs_data.get_subject_ids()
        else:
            fs_subject_list = subject_list

        long_combat_df = pd.read_csv(meta_data_path + "ukb_brain_imaging_metadata.csv")
        long_combat_df = long_combat_df[long_combat_df['eid'].isin(fs_subject_list)]
        long_combat_df = long_combat_df.rename(
            columns={"eid": "Subject_ID", "31-0.0": "Sex", "34-0.0": "Year_of_Birth", "53-2.0": "Date_of_Brain_Scan",
                     "54-2.0": "Assesment_Centre"})
        long_combat_df['Assesment_Centre'] = long_combat_df['Assesment_Centre'].apply(lambda x:
                                                                                      harmonize_fs_data.get_center(x))
        long_combat_df = long_combat_df.reset_index(drop=True)
        long_combat_df['Date_of_Brain_Scan'] = pd.to_datetime(long_combat_df['Date_of_Brain_Scan'])
        long_combat_df['Sex'] = long_combat_df['Sex'].apply(lambda x: harmonize_fs_data.ukb_subject_sex(x))
        long_combat_df['Age'] = long_combat_df['Year_of_Birth'].apply(lambda x: harmonize_fs_data.ukb_subject_age(x))
        long_combat_df['Age_at_Scan'] = long_combat_df.apply(lambda x: harmonize_fs_data.ukb_subject_age_at_scan(x),
                                                             axis=1)

        fs_features_df = pd.read_csv(meta_data_path + "ukb_fs_features.csv")
        fs_features_df = fs_features_df[fs_features_df['Subject_ID'].isin(fs_subject_list)]

        long_combat_df['Subject_ID'] = long_combat_df['Subject_ID'].astype(str)
        fs_features_df['Subject_ID'] = fs_features_df['Subject_ID'].astype(str)

        final_df = pd.merge(long_combat_df, fs_features_df, on='Subject_ID', how='outer')

        return final_df



import numpy as np
import random
from pandas_plink import read_plink1_bin
import pickle
genetic_data_path = "/ocean/projects/asc170022p/shared/Data/ukBiobank/UKB_Genetic_Data"
meta_dath_path = "/ocean/projects/asc170022p/shared/Data/ukBiobank/meta_data_files/"


class genetic_data_handler:
    """
    A class to represent family of methods that can fetch list of xarray objects based on combination of list of
    subjects and variants provided

    """

    def __init__(self):

        genetic_df = read_plink1_bin(genetic_data_path + "bed_files/" + "ukb22418_c" + "1" + "_b0_v2.bed",
                                     genetic_data_path + "bim_files/" + "ukb_snp_chr" + "1" + "_v2.bim",
                                     genetic_data_path + "fam_files/" + "ukb22418_c" + "1" + "_b0_v2_s488176.fam",
                                     verbose=True)
        genetic_df = genetic_df.set_index({"variant": "snp"})
        self.all_ukb_subjects = genetic_df.sample.values

    def get_subject_ids(self):
        """
        A utility function which lets user fetch numpy array containing the list of all subject ids
        Parameters:
        No parameter required
        Returns:
        categories list: A numpy array object
        :rtype: np.ndarray
        """
        if self.all_ukb_subjects is None:
            print("Please initialize Module")
            return None

        else:
            return self.all_ukb_subjects

    @staticmethod
    def get_path_genetic_files():
        """
        A static function which lets user see all the genetic data locations

        Parameters:
        No parameter required

        Returns:
        categories list: A string
        :rtype: str
        """

        return genetic_data_path

    def get_genetic_data(self, subject_list=None, variant_dict=None):
        """ The Primary function of the module which allows user to provide a subject list and dictionary which has keys
            as chromosome numbers and values as a list of variants of interest. If no subjects list is provided 100
            random subjects are considered.If no dictionary is provided as input, an exception is raised.

                Parameters:
                subject_list (list str): List of subjects which are of interest
                variant_dict (dict): A python Dictionary having keys as relevant chromosome number and values as list of
                variants that are of interest to us.

                Returns:
                list: A python list of xarrayDatasets

        """
        if subject_list is None:

            subject_list = random.sample(self.all_ukb_subjects, 100)

        elif variant_dict is None:
            raise Exception

        genetic_data_array = []
        for chr_num in variant_dict.keys():

            genetic_df = read_plink1_bin(genetic_data_path + "bed_files/" + "ukb22418_c" + chr_num + "_b0_v2.bed",
                                         genetic_data_path + "bim_files/" + "ukb_snp_chr" + chr_num + "_v2.bim",
                                         genetic_data_path + "fam_files/" + "ukb22418_c" + chr_num +
                                         "_b0_v2_s488176.fam", verbose=True)

            genetic_df = genetic_df.set_index({"variant": "snp"})
            genetic_df = genetic_df.where(genetic_df['sample'].isin(subject_list) & genetic_df['variant'].isin(
                variant_dict[chr_num]), drop=True)

            genetic_data_array.append(genetic_df)

        return genetic_data_array

    def get_genetic_data_from_snps(self, subject_list=None, snp_list = None):
        """ The Primary function of the module which allows user to provide a subject list and list of snps

                Parameters:

                subject_list (list str): List of subjects which are of interest
                SNP (list str): A python list containing snps of interest

                Returns:
                list: A python list of xarrayDatasets

        """
        if subject_list is None:

            subject_list = random.sample(self.all_ukb_subjects, 100)

        elif snp_list is None:
            raise Exception

        with open(meta_dath_path+'ukb_chr_variant_mapping.pkl', 'rb') as f:
            mapping_dict = pickle.load(f)

        variant_dict = {}
        genetic_data_array = []
        for key in mapping_dict.keys():
            res = mapping_dict[key].intersection(set(snp_list))

            if len(res) != 0:
                variant_dict[key] = list(res)

        for chr_num in variant_dict.keys():

            genetic_df = read_plink1_bin(genetic_data_path + "bed_files/" + "ukb22418_c" + chr_num + "_b0_v2.bed",
                                         genetic_data_path + "bim_files/" + "ukb_snp_chr" + chr_num + "_v2.bim",
                                         genetic_data_path + "fam_files/" + "ukb22418_c" + chr_num +
                                         "_b0_v2_s488176.fam", verbose=True)

            genetic_df = genetic_df.set_index({"variant": "snp"})
            genetic_df = genetic_df.where(genetic_df['sample'].isin(subject_list) & genetic_df['variant'].isin(
                variant_dict[chr_num]), drop=True)

            genetic_data_array.append(genetic_df)

        return genetic_data_array




import pandas as pd

meta_data_path = "/ocean/projects/asc170022p/shared/Data/ukBiobank/meta_data_files/"
#local_static_path = "/media/tighu/extended_storage/local_static_files_ukb/"

encoding_dict = {"ethnicity": {"-3.0": "Prefer not to answer", "-1.0": "Do not know", "1.0": "White", "2.0": "Mixed",
                               "3.0": "Asian or Asian British", "4.0": "Black or Black British", "5.0": "Chinese",
                               "6.0": "Other ethnic group", "1001.0": "British", "1002.0": "Irish",
                               "1003.0": "Any other white background", "2001.0": "White and Black Caribbean",
                               "2002.0": "White and Black African", "2003.0": "White and Asian",
                               "2004.0": "Any other mixed background", "3001.0": "Indian", "3002.0": "Pakistani",
                               "3003.0": "Bangladeshi", "3004.0": "Any other Asian background", "4001.0": "Caribbean",
                               "4002.0": "African", "4003.0": "Any other Black background"},
                 "nationality_encoding_dict": {},
                 "imaging_data_dictionary": {}, }


class scalar_data_handler:

    """
    A class to represent family of methods for handling and fetching scalar data

    ...

    Attributes
    ----------
    columns_to_read_for_field_id :
        list of relevant columns to read from metadata file


    Methods
    -------
    display_all_ukb_categories():
        displays all the categories

    get_field_ids_for_category(category_name):
        retrieves filed ids for a category

    get_subject_list_field_ids(category_name):
        retrieves list of subject relevant to that field id

    get_data_scalar(field_id,subject_id):
        fetches scalar data from metadata file

    """

    def __init__(self):
        self.columns_to_read_for_field_id = []
        return

    def display_all_ukb_categories(self):

        """
        The first prominent function which lets user see all the major categories present in the UKB dataset. Read
        categories from a static resource file inside package

        Parameters:
        Needs no parameters to operate

        Returns:
        categories list: List having all the unique categories of UKB
        :rtype: list of strings
        """
        unique_category_file_object = open(meta_data_path+"all_unique_categories.txt", "r")
        all_categories_list = unique_category_file_object.readlines()
        formatted_category_list = [a.rstrip() for a in all_categories_list]
        return formatted_category_list

    def get_field_ids_for_category(self, category_name=None):
        """
        A utility function which lets user see the associated field ids for each category, reads from a static file
        containing all the  category names and field ids

        Parameters:
        a string representing a category name

        Returns:
        categories list: pandas object having related field ids of the category
        :rtype: pandas table
        """

        temp_field_ids_df = pd.read_csv(meta_data_path+"ukb_field_ids.csv")
        return temp_field_ids_df[temp_field_ids_df['Category'] == category_name]

    def get_subject_list_field_ids(self, field_id=None):

        """
        A helper function which lets user see subjects having data related to a field_id. Reads a metadata file
        into a pandas object, only reading relevant columns because of size of the csv file

        Parameters:
        a int representing a field id

        Returns:
        subject list: a list of subject ids
        :rtype: list
        """

        with open(meta_data_path+'new_column_list_1.txt') as f:
            columns_list_df1 = f.readlines()

        for column_name in columns_list_df1:
            if int(column_name.split("-")[0]) == field_id:
                self.columns_to_read_for_field_id.append(column_name.strip())

        tempdf = pd.read_csv(meta_data_path+"ukb49570.csv", usecols=self.columns_to_read_for_field_id+['eid'])

        return tempdf['eid'].unique()

    def get_data_scalar(self, field_id=None, subject_list=None):
        """
        A helper function which lets user fetch scalar data associated with a field id and subject list.
        The fields in UKB dataset have been encoded in the following way: field_id-instance_number.array_index
        field_id : Numerical Number representing a particular feature of UKB dataset. For example 20252 represents
                   T1 NIFTI files
        instance_number: Currently 4 instances are present in the UKB dataset represents in which years of study was
                         data collected or recorded
                         Instance 0--> 2006-2010
                         Instance 1--> 2012-2013
                         Instance 2--> (2014+)
                         Instance 3--> (2019+)
        array_index: is the identifier present in cases where a single field id can have multiple values. An example
                     could be multiple readings of blood-pressure taken after fixed intervals

        Parameters:
        integers representing field id and subject id

        Returns:
        int/string: value of scalar data
        :rtype: int/str
        """

        field_id = str(field_id)
        cols_to_read = []
        for cols in self.columns_to_read_for_field_id:
            if cols.startswith(field_id):
                cols_to_read.append(cols)

        tempdf = pd.read_csv(meta_data_path+"ukb49570.csv", usecols=cols_to_read+['eid'])
        tempdf = tempdf.set_index('eid')

        return tempdf[tempdf.index.isin(subject_list)]
